from contextlib import contextmanager
from contextlib import suppress
from os import remove
from os.path import join
from tempfile import gettempdir
from typing import Generator
from typing import Optional
from uuid import uuid4

from opwen_email_server.utils.path import get_extension


def create_tempfilename(suffix: Optional[str] = None) -> str:
    extension = get_extension(suffix)
    return join(gettempdir(), f'{uuid4()}{extension}')


@contextmanager
def removing(path: str) -> Generator[str, None, None]:
    try:
        yield path
    finally:
        remove_if_exists(path)


def remove_if_exists(path: str):
    with suppress(FileNotFoundError):
        remove(path)



from pathlib import Path
from subprocess import check_call  # nosec
from subprocess import check_output  # nosec


def _find_device(stdout: bytes, device_id: str):
    for line in stdout.splitlines():
        line = line.decode('utf-8')
        if 'Huawei' not in line:
            continue
        if device_id in line:
            return True
    return False


def modem_is_plugged(modem=None):
    result = check_output('/usr/bin/lsusb', shell=True)  # nosec
    return _find_device(result, modem.uid if modem else '12d1:')


def modem_is_setup(target_mode: str):
    result = check_output('/usr/bin/lsusb', shell=True)  # nosec
    return _find_device(result, '12d1:{}'.format(target_mode))


def setup_modem(config: Path):
    check_call('/usr/sbin/usb_modeswitch --config-file="{}"'.format(config.absolute()), shell=True)  # nosec



from typing import Callable
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional

from libcloud.storage.types import ObjectDoesNotExistError
from requests import RequestException
from requests import post as http_post

from opwen_email_server.constants import events
from opwen_email_server.constants import github
from opwen_email_server.services.storage import AzureObjectStorage
from opwen_email_server.utils.log import LogMixin


class AnyOfBasicAuth(LogMixin):
    def __init__(self, auths: Iterable[Callable[[str, str, Optional[List[str]]], Optional[Dict[str, str]]]]):
        self._auths = list(auths)

    def __call__(self, username, password, required_scopes=None):
        for auth in self._auths:
            user = auth(username, password, required_scopes)
            if user is not None:
                return user

        return None


class BasicAuth(LogMixin):
    def __init__(self, users: dict):
        self._users = dict(users)

    def __call__(self, username, password, required_scopes=None):
        if not username or not password:
            return None

        try:
            user = self._users[username]
        except KeyError:
            self.log_event(events.UNKNOWN_USER, {'username': username})  # noqa: E501  # yapf: disable
            return None

        if user['password'] != password:
            self.log_event(events.BAD_PASSWORD, {'username': username})  # noqa: E501  # yapf: disable
            return None

        if not self._has_scopes(user, required_scopes):
            self.log_event(events.MISSING_SCOPES, {'username': username})  # noqa: E501  # yapf: disable
            return None

        return {'sub': username}

    @classmethod
    def _has_scopes(cls, user, required_scopes) -> bool:
        if not required_scopes:
            return True

        return set(required_scopes).issubset(user.get('scopes', []))


class GithubBasicAuth(LogMixin):
    def __init__(self, organization: str, team: str, page_size: int = 50):
        self._organization = organization
        self._team = team
        self._page_size = page_size

    def __call__(self, username, password, required_scopes=None):
        if not username or not password or not self._organization or not self._team:
            return None

        try:
            team_members = self._fetch_team_members(access_token=password)
            user_exists = any(username == team_member for team_member in team_members)
        except RequestException:
            self.log_event(events.BAD_PASSWORD, {'username': username})  # noqa: E501  # yapf: disable
            return None

        if not user_exists:
            self.log_event(events.UNKNOWN_USER, {'username': username})  # noqa: E501  # yapf: disable
            return None

        return {'sub': username}

    def _fetch_team_members(self, access_token: str) -> Iterable[str]:
        cursor = None

        while True:
            response = http_post(
                url=github.GRAPHQL_URL,
                json={
                    'query': '''
                        query($organization:String!, $team:String!, $cursor:String, $first:Int!) {
                            organization(login:$organization) {
                                team(slug:$team) {
                                    members(after:$cursor, first:$first, orderBy:{ field:LOGIN, direction:DESC }) {
                                        edges {
                                            cursor
                                        }
                                        nodes {
                                            login
                                        }
                                    }
                                }
                            }
                        }
                    ''',
                    'variables': {
                        'organization': self._organization,
                        'team': self._team,
                        'cursor': cursor,
                        'first': self._page_size,
                    },
                },
                headers={
                    'Authorization': f'Bearer {access_token}',
                },
            )
            response.raise_for_status()

            members = response.json()['data']['organization']['team']['members']
            nodes = members['nodes']
            edges = members['edges']

            for member in nodes:
                yield member['login']

            if len(nodes) < self._page_size:
                break

            cursor = edges[-1]['cursor']


class AzureAuth(LogMixin):
    def __init__(self, storage: AzureObjectStorage) -> None:
        self._storage = storage

    def insert(self, client_id: str, domain: str, owner: str):
        auth = {'client_id': client_id, 'owner': owner, 'domain': domain}
        self._storage.store_object(self._client_id_file(client_id), auth)
        self._storage.store_object(self._domain_file(domain), auth)
        self.log_info('Registered client %s at domain %s', client_id, domain)

    def is_owner(self, domain: str, username: str) -> bool:
        try:
            auth = self._storage.fetch_object(self._domain_file(domain))
        except ObjectDoesNotExistError:
            self.log_warning('Unrecognized domain %s', domain)
            return False

        return auth.get('owner') == username

    def delete(self, client_id: str, domain: str) -> bool:
        self._storage.delete(self._domain_file(domain))
        self._storage.delete(self._client_id_file(client_id))
        return True

    def client_id_for(self, domain: str) -> Optional[str]:
        try:
            auth = self._storage.fetch_object(self._domain_file(domain))
        except ObjectDoesNotExistError:
            self.log_warning('Unrecognized domain %s', domain)
            return None

        client_id = auth['client_id']

        self.log_debug('Domain %s has client %s', domain, client_id)
        return client_id

    def domain_for(self, client_id: str) -> Optional[str]:
        try:
            auth = self._storage.fetch_object(self._client_id_file(client_id))
        except ObjectDoesNotExistError:
            self.log_warning('Unrecognized client %s', client_id)
            return None
        else:
            domain = auth['domain']
            self.log_debug('Client %s has domain %s', client_id, domain)
            return domain

    def domains(self) -> Iterable[str]:
        return self._storage.iter(self._domain_file(''))

    @classmethod
    def _domain_file(cls, domain: str) -> str:
        return f'domain/{domain}'

    @classmethod
    def _client_id_file(cls, client_id: str) -> str:
        return f'client_id/{client_id}'


class NoAuth(LogMixin):
    def is_owner(self, domain: str, username: str) -> bool:
        return True

    def domain_for(self, client_id: str) -> str:
        return 'service'



try:
    from html import unescape
except ImportError:
    from html.parser import HTMLParser
    unescape = HTMLParser().unescape  # type: ignore

from re import IGNORECASE
from re import compile as re_compile
from typing import Optional

from crontab import CronItem
from wtforms import StringField
from wtforms import TextAreaField
from wtforms.validators import HostnameValidation
from wtforms.validators import Regexp
from wtforms.validators import ValidationError


class CronSchedule:
    def __init__(self, message=None):
        self.message = message

    def __call__(self, form, field, message=None):
        schedule = (field.data or '').strip()
        if not schedule:
            return

        try:
            CronItem().setall(schedule)
        except (KeyError, ValueError):
            message = message or self.message or field.gettext('Invalid cron')
            raise ValidationError(message)


class Emails(Regexp):
    def __init__(self, email_address_delimiter, message=None):
        self.validate_hostname = HostnameValidation(require_tld=True)
        self.email_address_delimiter = email_address_delimiter
        super().__init__(r'^.+@([^.@][^@]+)$', IGNORECASE, message)

    # noinspection PyMethodOverriding
    def __call__(self, form, field):
        message = self.message
        if message is None:
            message = field.gettext('Invalid email address.')

        messages = message.split(self.email_address_delimiter)
        for message in messages:
            match = super().__call__(form, field, message.strip())
            if not self.validate_hostname(match.group(1)):
                raise ValidationError(message)


class HtmlTextAreaField(TextAreaField):
    _script_tags_re = re_compile(r'<script[\s\S]+?/script>', IGNORECASE)

    # noinspection PyAttributeOutsideInit
    def process_formdata(self, valuelist):
        super().process_formdata(valuelist)
        self.data = self._to_safe_html(self.data)

    def _value(self):
        value = super()._value()
        return self._to_safe_html(value)

    @classmethod
    def _remove_script_tags(cls, data: str) -> str:
        match = cls._script_tags_re.search(data)
        while match:
            start = match.start()
            end = match.end()
            data = data[:start] + data[end:]
            match = cls._script_tags_re.search(data)

        return data

    @classmethod
    def _to_safe_html(cls, data: Optional[str]) -> str:
        if not data:
            return ''

        data = unescape(data)
        data = cls._remove_script_tags(data)
        return data


class SuffixedStringField(StringField):
    def __init__(self, suffix: str = '', *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._suffix = suffix

    # noinspection PyAttributeOutsideInit
    def process_formdata(self, valuelist):
        super().process_formdata(valuelist)
        self.data = self._ensure_suffix(self.data)

    def _value(self):
        value = super()._value()
        return self._ensure_suffix(value)

    def _ensure_suffix(self, value: str) -> str:
        if not value:
            return value

        if not value.endswith(self._suffix):
            value = value + self._suffix

        return value.lower()



from logging import CRITICAL
from logging import DEBUG
from logging import INFO
from logging import WARNING
from logging import Logger
from logging import getLogger
from typing import Any
from typing import Iterable
from typing import Optional

from applicationinsights import TelemetryClient
from applicationinsights.channel import AsynchronousQueue
from applicationinsights.channel import AsynchronousSender
from applicationinsights.channel import NullSender
from applicationinsights.channel import TelemetryChannel
from cached_property import cached_property

from opwen_email_server.config import APPINSIGHTS_HOST
from opwen_email_server.config import APPINSIGHTS_KEY
from opwen_email_server.utils.collections import append
from opwen_email_server.utils.collections import singleton


@singleton
def _create_telemetry_channel() -> TelemetryChannel:
    sender = AsynchronousSender(APPINSIGHTS_HOST) if APPINSIGHTS_HOST else NullSender()
    queue = AsynchronousQueue(sender)
    return TelemetryChannel(queue=queue)


class LogMixin:
    _telemetry_channel = _create_telemetry_channel()
    _telemetry_key = APPINSIGHTS_KEY or '00000000-0000-0000-0000-000000000000'

    @cached_property
    def _logger(self) -> Logger:
        return getLogger('gunicorn.error')

    @cached_property
    def _telemetry_client(self) -> TelemetryClient:
        return TelemetryClient(self._telemetry_key, self._telemetry_channel)

    def log_debug(self, message: str, *args: Any):
        self._log(DEBUG, message, args)

    def log_info(self, message: str, *args: Any):
        self._log(INFO, message, args)

    def log_warning(self, message: str, *args: Any):
        self._log(WARNING, message, args)

    def log_exception(self, ex: Exception, message: str, *args: Any):
        self._log(CRITICAL, message + ' (%r)', append(args, ex))

        # noinspection PyBroadException
        try:
            raise ex
        except Exception:
            self._telemetry_client.track_exception()
            self._telemetry_channel.flush()

    def _log(self, level: int, log_message: str, log_args: Iterable[Any]):
        self._logger.log(level, log_message, *log_args)

    def log_event(self, event_name: str, properties: Optional[dict] = None):
        self.log_info('%s|%s', event_name, properties)
        self._telemetry_client.track_event(event_name, properties)



from celery import Celery
from celery.schedules import crontab
from celery.utils.log import get_task_logger

from opwen_email_client.webapp.actions import RestartApp
from opwen_email_client.webapp.actions import StartInternetConnection
from opwen_email_client.webapp.actions import SyncEmails
from opwen_email_client.webapp.actions import UpdateCode
from opwen_email_client.webapp.config import AppConfig

app = Celery(__name__, broker=AppConfig.CELERY_BROKER_URL)
app.conf.beat_schedule_filename = AppConfig.CELERY_BEAT_SCHEDULE_FILENAME
log = get_task_logger(__name__)


# noinspection PyUnusedLocal
@app.on_after_configure.connect
def setup_periodic_tasks(sender, **kwargs):
    sync_schedule = AppConfig.SYNC_SCHEDULE.split(' ')
    if len(sync_schedule) == 5:
        sender.add_periodic_task(
            schedule=crontab(*sync_schedule),
            sig=sync.s('periodic_email_sync'),
        )


# noinspection PyUnusedLocal
@app.task(ignore_result=True)
def sync(*args, **kwargs):
    from opwen_email_client.webapp import app as webapp

    sync_emails = SyncEmails(
        log=log,
        email_sync=webapp.ioc.email_sync,
        email_store=webapp.ioc.email_store,
        user_store=webapp.ioc.user_store,
    )

    start_internet_connection = StartInternetConnection(
        state_dir=AppConfig.STATE_BASEDIR,
        modem_config_dir=AppConfig.MODEM_CONFIG_DIR,
        sim_config_dir=AppConfig.SIM_CONFIG_DIR,
        sim_type=AppConfig.SIM_TYPE,
    )

    if AppConfig.SIM_TYPE != 'LocalOnly':
        with start_internet_connection():
            sync_emails()


# noinspection PyUnusedLocal
@app.task(ignore_result=True)
def update(*args, **kwargs):
    update_code = UpdateCode(
        version=kwargs.get('version') or '',
        log=log,
    )

    restart_app = RestartApp(restart_paths=AppConfig.RESTART_PATHS)

    start_internet_connection = StartInternetConnection(
        state_dir=AppConfig.STATE_BASEDIR,
        modem_config_dir=AppConfig.MODEM_CONFIG_DIR,
        sim_config_dir=AppConfig.SIM_CONFIG_DIR,
        sim_type=AppConfig.SIM_TYPE,
    )

    if AppConfig.SIM_TYPE != 'LocalOnly':
        with start_internet_connection():
            update_code()

        restart_app()



from abc import ABCMeta
from abc import abstractmethod
from base64 import b64decode
from base64 import b64encode


class AttachmentEncoder(metaclass=ABCMeta):
    @abstractmethod
    def encode(self, content: bytes) -> str:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def decode(self, encoded: str) -> bytes:
        raise NotImplementedError  # pragma: no cover


class Base64AttachmentEncoder(AttachmentEncoder):
    encoding = 'utf-8'

    def encode(self, content):
        content_bytes = b64encode(content)
        return content_bytes.decode(self.encoding)

    def decode(self, encoded):
        content_bytes = encoded.encode(self.encoding)
        return b64decode(content_bytes)



from fileinput import input as fileinput
from gzip import GzipFile
from os import listdir
from os.path import isdir
from os.path import join
from pathlib import Path
from typing import Callable
from typing import Iterable
from typing import Optional
from typing import Union


def subdirectories(root: str) -> Iterable[str]:
    try:
        return (sub for sub in listdir(root) if isdir(join(root, sub)))
    except OSError:
        return []


def replace_line(path: str, match: Callable[[str], bool], replacement: str):
    for line in fileinput(path, inplace=True):
        if match(line):
            end = ''
            if line.endswith('\r\n'):
                end = '\r\n'
            elif line.endswith('\n'):
                end = '\n'
            if replacement.endswith(end):
                end = ''
            print(replacement, end=end)
        else:
            print(line, end='')


def backup(path: Union[str, Path], suffix: str = '.bak.gz') -> Optional[Path]:
    path = Path(path)

    if not path.is_file():
        return None

    backup_path = '{}{}'.format(path, suffix)
    with GzipFile(backup_path, mode='ab') as fout:
        with path.open('rb') as fin:
            for line in fin:
                fout.write(line)

    return Path(backup_path)



from abc import ABCMeta
from abc import abstractmethod
from os import getenv
from os import path
from urllib.parse import urlencode

from requests import get as http_get
from requests import post as http_post


class EmailServerClient(metaclass=ABCMeta):
    @abstractmethod
    def upload(self, resource_id: str, container: str):
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def download(self) -> str:
        raise NotImplementedError  # pragma: no cover


class HttpEmailServerClient(EmailServerClient):
    def __init__(self, compression: str, endpoint: str, client_id: str):
        self._compression = compression
        self._endpoint = endpoint
        self._client_id = client_id

    @property
    def _base_url(self) -> str:
        return '{endpoint}/api/email'.format(endpoint=self._endpoint)

    @property
    def _upload_url(self) -> str:
        return '{base_url}/upload/{client_id}'.format(
            base_url=self._base_url,
            client_id=self._client_id,
        )

    @property
    def _download_url(self) -> str:
        return '{base_url}/download/{client_id}?{query}'.format(
            base_url=self._base_url,
            client_id=self._client_id,
            query=urlencode({
                'compression': self._compression,
            }),
        )

    def upload(self, resource_id, container):
        payload = {
            'resource_id': resource_id,
        }

        response = http_post(self._upload_url, json=payload)
        response.raise_for_status()

    def download(self):
        response = http_get(self._download_url)
        response.raise_for_status()
        resource_id = response.json()['resource_id']

        return resource_id


class LocalEmailServerClient(EmailServerClient):
    def download(self) -> str:
        root = getenv('OPWEN_REMOTE_ACCOUNT_NAME')
        container = getenv('OPWEN_REMOTE_RESOURCE_CONTAINER')
        resource_id = 'sync.tar.gz'
        local_file = path.join(root, container, resource_id)
        if not path.isfile(local_file):
            return ''
        return resource_id

    def upload(self, resource_id: str, container: str):
        print('Uploaded {}/{}'.format(container, resource_id))
